"""Enum with common status for CompilerTasks SimulatorTasks."""

from enum import Enum


class TaskStatus(str, Enum):
    """Contains allowed statuses for Task objects.

    Args:
        Enum (enum): Class inherits from Enum
    """

    PENDING = "pending"
    PREPARING = "preparing"
    EXECUTING = "executing"
    POSTPROCESSING = "postprocessing"
    FINISHED = "finished"
    ERROR = "error"
    ABORTED = "aborted"



# generated by fastapi-codegen:
#   filename:  ./api-wrapper/openapi.yaml
#   timestamp: 2025-01-08T11:10:30+00:00

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Extra, Field, PositiveFloat, PositiveInt, conint, constr

class CompilerStatus(Enum):
    running = 'running'
    offline = 'offline'
    maintenance = 'maintenance'


class CompilerTable(BaseModel):
    name: str = Field(..., title='Name')
    manufacturer: Optional[str] = Field(..., title='Manufacturer')
    status: CompilerStatus
    description: Optional[constr(min_length=0, max_length=512)] = Field(
        ..., title='Description'
    )
    time_limit: PositiveFloat = Field(..., title='Time Limit')
    passes_available: Optional[Dict[str, bool]] = Field(None, title='Passes Available')
    compiler_id: int = Field(..., title='Compiler Id')
#    compiler_alias: str = Field(..., title='Compiler Alias')

class CompilerTaskBase(BaseModel):
    compiler_id: Optional[int] = Field(..., title='Compiler Id')
    compilation_pass: str
    options: Optional[Dict[str, Any]] = Field(None, title='Options')


class CompilerTaskResult(BaseModel):
    meta_data: Any = Field(..., title='Meta Data')
    status: str = Field(..., title='Status')
    status_message: Optional[str] = Field(..., title='Status Message')
    start_date_time: Optional[str] = Field(..., title='Start Date Time')
    end_date_time: Optional[str] = Field(..., title='End Date Time')
    total_runtime: Optional[float] = Field(..., title='Total Runtime')
    compiled_circuit: Optional[Union[str,List[str]]] = Field(..., title='Compiled Circuit')


class Primitive(Enum):
    sampling = 'sampling'
    quantum_state = 'quantum_state'


class QPUOptions(BaseModel):
    class Config:
        extra = Extra.forbid

    coupling_map: Optional[List[List[conint(ge=0)]]] = Field(None, title='Coupling Map')
    basis_gates: Optional[List[str]] = Field(None, title='Basis Gates')


class QPUStatus(Enum):
    idle = 'idle'
    running = 'running'
    offline = 'offline'
    maintenance = 'maintenance'


class QPUTaskBase(BaseModel):
    total_shots: Optional[conint(ge=1, le=1000000)] = Field(1, title='Total Shots')
    qpu_id: int = Field(..., title='Qpu Id')
    primitive: Optional[Primitive] = 'sampling'
    qpu_options: Optional[QPUOptions]
    class Config:  
        use_enum_values = True


class QPUTaskResult(BaseModel):
    meta_data: Any = Field(..., title='Meta Data')
    status: str = Field(..., title='Status')
    status_message: Optional[str] = Field(..., title='Status Message')
    start_date_time: Optional[str] = Field(..., title='Start Date Time')
    end_date_time: Optional[str] = Field(..., title='End Date Time')
    total_runtime: Optional[float] = Field(..., title='Total Runtime')
    data: Optional[str] = Field(..., title='Data')


class Technology(Enum):
    statevector_simulator = 'statevector_simulator'
    trapped_ions = 'trapped_ions'
    neutral_atoms = 'neutral_atoms'
    vacancy_centers = 'vacancy_centers'


class ValidationError(BaseModel):
    loc: List[Union[str, int]] = Field(..., title='Location')
    msg: str = Field(..., title='Message')
    type: str = Field(..., title='Error Type')


class HTTPValidationError(BaseModel):
    detail: Optional[List[ValidationError]] = Field(None, title='Detail')


class JobInfo(BaseModel):
    circuit: str = Field(..., title='Circuit')
    primitive: Optional[Primitive] = 'sampling'
    name: constr(min_length=1, max_length=64) = Field(..., title='Name')
    comment: Optional[constr(min_length=0, max_length=512)] = Field(
        ..., title='Comment'
    )
    job_id: UUID = Field(..., title='Job Id')
    submission_date_time: datetime = Field(..., title='Submission Date Time')
    status: str = Field(..., title='Status')
    status_message: Optional[str] = Field(..., title='Status Message')
    qpu_name: str = Field(..., title='Qpu Name')
    last_compiler_task: Optional[CompilerTaskBase]
    last_qpu_task: Optional[QPUTaskBase]


class JobResult(BaseModel):
    status: str = Field(..., title='Status')
    status_message: Optional[str] = Field(..., title='Status Message')
    start_date_time: Optional[str] = Field(..., title='Start Date Time')
    end_date_time: Optional[str] = Field(..., title='End Date Time')
    total_runtime: Optional[float] = Field(..., title='Total Runtime')
    last_compiler_result: Optional[CompilerTaskResult]
    last_qpu_result: Optional[QPUTaskResult]


class JobSubmission(BaseModel):
    circuit: str = Field(..., title='Circuit')
    primitive: Optional[Primitive] = 'sampling'
    name: constr(min_length=1, max_length=64) = Field(..., title='Name')
    comment: Optional[constr(min_length=0, max_length=512)] = Field(
        ..., title='Comment'
    )
    tasks: List[Union[CompilerTaskBase, QPUTaskBase]] = Field(..., title='Tasks')
    class Config:  
        use_enum_values = True


class QPUTable(BaseModel):
    manufacturer: str = Field(..., title='Manufacturer')
    name: str = Field(..., title='Name')
    technology: Technology
    status: QPUStatus
    number_of_qubits_available: PositiveInt = Field(
        ..., title='Number Of Qubits Available'
    )
    circuit_depth_limit: conint(ge=0) = Field(..., title='Circuit Depth Limit')
    shot_limit: PositiveInt = Field(..., title='Shot Limit')
    description: Optional[constr(min_length=0, max_length=512)] = Field(
        ..., title='Description'
    )
    time_limit: PositiveFloat = Field(..., title='Time Limit')
    measurement_required: Optional[bool] = Field(..., title='Measurement Required')
    qpu_id: int = Field(..., title='Qpu Id')
    qpu_alias: str = Field(..., title='Qpu Alias')



"""Module for handling results from QciConnectClient."""
import contextlib
import json

import numpy as np
from hequate_common.models import CompilerTaskResult, QPUTaskResult

from .client import QciConnectClient
from .exceptions import ResultError
from .utils import timestamp_to_datetime


class CompilerResult:
    """Result of a compile job.

    Attributes:
        _compiler_task_result: Result as it is returned by the platform parsed by pydantic.
        compiled_circuits: Workaround - we want to return the circuits as a list.
    """

    _compiler_task_result: CompilerTaskResult
    compiled_circuits: list

    def __getattr__(self, attribute):
        """Redirects retrieval of attributes to _compiler_task_result and hides pydantic noise.

        Args:
            attribute: Name of attribute to get.

        Returns: attribute.
        """
        if attribute != "compiled_circuits":
            return getattr(self._compiler_task_result, attribute)
        else:
            return self.compiled_circuits

    def __dir__(self):
        """Fixes autocompletion and hides pydantic noise."""
        dir_output = list(self._compiler_task_result.model_fields_set)
        dir_output.append("compiled_circuits")
        dir_output.remove("compiled_circuit")
        return dir_output

    @classmethod
    def from_compiler_task_result(cls, result: CompilerTaskResult):
        """Creates CompilerResult from CompilerTaskResult.

        Args:
            result: compiler task result to create compiler result from.

        Returns: Compiler result
        """
        compiler_result = cls()
        compiler_result._compiler_task_result = result
        if isinstance(result.compiled_circuit, list):
            compiler_result.compiled_circuits = result.compiled_circuit
        else:
            compiler_result.compiled_circuits = [result.compiled_circuit]
        return compiler_result


class BackendResult:
    """Result of a qpu job.

    Attributes:
        _qpu_task_result: Result as it is returned by the platform parsed by pydantic.
    """

    _qpu_task_result: QPUTaskResult

    def __getattr__(self, attribute):
        """Redirects retrieval of attributes to _qpu_task_result and hides pydantic noise.

        Args:
            attribute: Name of attribute to get.

        Returns: attribute.
        """
        return getattr(self._qpu_task_result, attribute)

    def __dir__(self):
        """Fixes autocompletion and hides pydantic noise."""
        return list(self._qpu_task_result.model_fields_set)

    @classmethod
    def from_qpu_task_result(cls, result: QPUTaskResult):
        """Creates BackendResult from QPUTaskResult.

        Args:
            result: QPU task result to create backend result from.

        Returns: backend result
        """
        backend_result = cls()
        backend_result._qpu_task_result = result

        if isinstance(backend_result._qpu_task_result.data, str):
            backend_result._qpu_task_result.data = json.loads(backend_result._qpu_task_result.data)

        if type(backend_result._qpu_task_result.data) is dict:
            # See whether the data is a dictionary of real valued numpy arrays
            # (e.g. for "sampling" primitive)
            with contextlib.suppress(Exception):
                backend_result._qpu_task_result.data = {
                    key: np.ndarray(value)
                    for key, value in backend_result._qpu_task_result.data.items()
                }

            # Otherwise, convert to complex numbers
            # (e.g. necessary for "quantum_state" primitive)
            with contextlib.suppress(Exception):
                backend_result._qpu_task_result.data = {
                    key: complex(value)
                    for key, value in backend_result._qpu_task_result.data.items()
                }

        elif type(backend_result._qpu_task_result.data) is list:
            with contextlib.suppress(Exception):
                backend_result._qpu_task_result.data = np.ndarray(
                    backend_result._qpu_task_result.data
                )

        else:
            raise ResultError(
                f"Unexpected data type of backend_result._qpu_task_result.data: "
                f"{type(backend_result._qpu_task_result.data)}"
            )

        try:
            backend_result._qpu_task_result.start_date_time = timestamp_to_datetime(
                backend_result._qpu_task_result.start_date_time
            )
            backend_result._qpu_task_result.end_date_time = timestamp_to_datetime(
                backend_result._qpu_task_result.end_date_time
            )
        except Exception: # noqa: BLE001
            pass

        return backend_result


class FutureResult:
    """Object that can be used to poll/retrieve the result of an job.

    Attributes:
        status: status of the job.
        result: Result as it is returned by the platform parsed by pydantic.
    """

    def __init__(self, job_id: str, client: QciConnectClient):
        """Initialize the FutureResult object."""
        self._client = client
        self._job_id = job_id
        self._result = None
        self.update()

    @property
    def job_id(self) -> str:
        """Returns the job ID of the job."""
        return self._job_id

    @property
    def status(self) -> str:
        """Returns the current status of the job. Updates status and result if necessary."""
        if not self._result:
            self.update()
        return self._status

    def update(self):
        """Updates status of the job and retrieves result if job has finished."""
        self._status = self._client.get_job_status(self._job_id)
        self._result = self._client.retrieve_result(self._job_id)


class FutureBackendResult(FutureResult):
    """Future result for backend jobs."""
    @property
    def result(self) -> BackendResult | None:
        """Updates future and returns BackendResult if ready.

        Returns: BackendResult object.
        """
        if not self._result:
            self.update()
        if self._result:
            return BackendResult.from_qpu_task_result(self._result.last_qpu_result)
        return None


class FutureCompilerResult(FutureResult):
    """Future result for compiler tasks."""
    @property
    def result(self) -> CompilerResult | None:
        """Updates future and returns CompilerResult if ready.

        Returns: BackendResult object.
        """
        if not self._result:
            self.update()
        if self._result:
            return CompilerResult.from_compiler_task_result(self._result.last_compiler_result)
        return None


class Results(dict):
    """Dictionary-like object for managing results from QCI Connect server."""
    def __init__(self, client: QciConnectClient):
        """Initialize the Results object with a QciConnectClient instance."""
        super().__init__(self)
        self._client = client

    def __dir__(self):
        """Update the results dictionary with the latest jobs from the server."""
        self._update_results()
        return super().__dir__()

    def _update_results(self):
        super().clear()
        jobs = self._client.get_jobs()
        for job in jobs:
            self._update_result(job)

    def _update_result(self, job):
        if job["status"] == "SUCCESS":
            job_id = job["job_id"]
            if job["last_compiler_task"] is not None:
                super().__setitem__(job_id, CompilerResult)
            if job["last_qpu_task"] is not None:
                super().__setitem__(job_id, BackendResult)

    def __getitem__(self, key: str) -> BackendResult | CompilerResult | None:
        """Get the result for a given job ID."""
        if key not in self:
            job = self._client.get_job(key)
            job["job_id"] = key
            self._update_result(job)
        if key not in self:
            return None

        if super().__getitem__(key) == BackendResult:
            super().__setitem__(
                key,
                BackendResult.from_qpu_task_result(
                    self._client.retrieve_result(key).last_qpu_result
                ),
            )
        elif super().__getitem__(key) == CompilerResult:
            super().__setitem__(
                key,
                CompilerResult.from_compiler_task_result(
                    self._client.retrieve_result(key).last_compiler_result
                ),
            )
        return super().__getitem__(key)

    def refresh(self):
        """Refresh the results dictionary with the latest results from the server."""
        self._update_results()



"""Provides utility functions for the qciconnect package."""
import datetime
import json
from pathlib import Path


def timestamp_to_datetime(timestamp):
    """Converts a timestamp string to a datetime object.

    Args:
       timestamp (str): The timestamp string in the format "%Y-%m-%dT%H:%M:%S.%f"
    """
    try:
        result = datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%f")
    except ValueError:
        result = timestamp
    return result


def read_json(file_path):
    """Reads a JSON file and returns the data."""
    with Path.open(file_path) as fp:
        return json.load(fp)



"""Provides helper methods to convert between JSON and Python data formats."""

from datetime import datetime, timedelta
from json import dumps, loads

import numpy as np
from numpy import asarray
from qciconnect_common.qciconnect_logging.logger import QCIConnectLogging

logger = QCIConnectLogging().get_logger()


class JSONNumpyConverter:
    """Provides helper methods to convert between JSON and numpy data formats."""

    @staticmethod
    def nparray_to_json_list(nparray: np.ndarray):
        """Convert NumPy array to JSON list.

        Args:
            nparray (np.array): NumPy array to be converted

        Returns:
            json_list: JSON list equivalent to nparray
        """
        python_list = nparray.tolist()
        json_list = dumps(python_list)
        return json_list

    @staticmethod
    def json_list_to_nparray(json_list: str):
        """Convert JSON list to NumPy array.

        Args:
            json_list (JSON list): JSON list to be converted

        Returns:
            np.array: NumPy array representation of json_list
        """
        python_list = loads(json_list)
        try:
            nparray = asarray(python_list)
        except ValueError:
            logger.error(
                """
                Invalid input. 
                Content or shape of input JSON list is not compatible with NumPy array 
                (e.g. inhomogenous shape or datatypes).
                """
            )
        except TypeError:
            logger.error("Invalid input. Input is not a (JSON) string.")
        else:
            return nparray

    @staticmethod
    def dict_to_json(dictionary: dict):
        """Convert Python dictionary to JSON object.

        Args:
            dictionary (dict): Dictionary to be converted

        Returns:
            JSON list: JSON list equivalent to dictionary
        """
        return dumps(dictionary)

    @staticmethod
    def json_to_dict(json_object: str):
        """Convert JSON object to Python dictionary.

        Args:
            json_object (JSON list): JSON list to be converted

        Returns:
            dict: Dictionary equivalent to json_object
        """
        return loads(json_object)

    @staticmethod
    def complex_python_to_json(complex_number: complex):
        """Convert Python complex number to string.

        Args:
            complex_number (complex): Complex number to be converted

        Returns:
            str: String representation of complex number.
            Example: a Python complex number 1.1+2.2j (or complex(1.1, 2.2) ) with real part 1.1
            and imaginary part 2.2 becomes '(1.1+2.2j)'.
        """
        if isinstance(complex_number, complex):
            json_complex_number = str(complex_number)
            return json_complex_number
        else:
            error = TypeError(f"Input {complex_number} is not a valid number of type complex.")
            logger.error(error)
            raise error

    @staticmethod
    def complex_json_to_python(json_complex_number: str):
        """Convert JSON string representation of complex number to Python complex.

        Args:
            json_complex_number (str): Complex number to be converted, must be of forms
            '(1.1+2.2j)', '1.1+2.2j', '2.2j + 1.1', '1.1', or '2.2j'.

        Returns:
            complex: Complex number in standard Python format.
        """
        try:
            complex_number = complex(json_complex_number)
        except ValueError:
            logger.error("Invalid input string. Cannot be converted to complex.")
        except TypeError:
            logger.error(
                """
                Invalid input format. 
                Must be string of form '(1.1+2.2j)', '1.1+2.2j', '2.2j + 1.1', '1.1', or '2.2j'.
                """
            )
        else:
            return complex_number

    @staticmethod
    def complex_nparray_to_json_list(complex_array: np.ndarray):
        """Convert NumPy array with complex numbers as entries to JSON list.

        Args:
            complex_array (np.array, dtype = complex): Complex NumPy array to be converted.

        Raises:
            error: TypeError if entries are not complex.
            error: TypeError if input is not a np.array.

        Returns:
            str: JSON list with string representations (e.g. '(1.1+2.2j)') of complex entries.
        """
        if isinstance(complex_array, np.ndarray):
            if (complex_array.dtype != complex) and (complex_array.dtype != "complex128"):
                error = TypeError(
                    f"Entries of {complex_array} are not dtype=complex, but {complex_array.dtype}."
                )
                logger.error(error)
                raise error
            else:
                complex_list = complex_array.tolist()
                json_list = str(complex_list)
                return json_list
        else:
            error = TypeError(f"Input {complex_array} is not a -NumPy array with dtype = complex.")
            logger.error(error)
            raise error


class JSONDateTimeConverter:
    """Provides helper methods to convert between JSON and Python datetime formats."""

    @staticmethod
    def datetime_to_json_datetime(datetime_object: datetime):
        """Converts Python datetime object to ISO JSON string.

        Args:
            datetime_object (datetime): Python datetime object to be converted,

        Raises:
            error: TypeError if input is not a datetime object.

        Returns:
            str: JSON string representation of datetime in ISO format with microsecond precison
            (e.g. "2020-02-02T01:02:03.000004Z").
        """
        if isinstance(datetime_object, datetime):
            datetime_json = datetime_object.strftime("%Y-%m-%dT%H:%M:%S.%f")
            return datetime_json
        else:
            error = TypeError(f"Input {datetime_object} is not a valid datetime object.")
            logger.error(error)
            raise error

    @staticmethod
    def json_datetime_to_datetime(datetime_str: str):
        """Converts Python datetime object to ISO JSON string.

        Args:
            datetime_object (datetime): Python datetime object to be converted,

        Raises:
            error: TypeError if input is not a datetime object.

        Returns:
            str: JSON string representation of datetime in ISO format with microsecond precison
            (e.g. "2020-02-02T01:02:03.000004Z").
        """
        if isinstance(datetime_str, str):
            datetime_object = datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S.%f")
            return datetime_object
        else:
            error = TypeError(f"Input {datetime_str} is not a valid datetime object.")
            logger.error(error)
            raise error


    @staticmethod
    def timedelta_to_json(timedelta_object: timedelta):
        """Converts Python datetime.timedelta object to JSON float (string) in seconds.

        Args:
            timedelta_object (datetime.timedelta): Python datetime.timedelta object to be
            converted.

        Raises:
            error: TypeError if input is not a datetime.timedelta object.

        Returns:
            str: JSON string (float) representation of datetime.timedelta in seconds.
        """
        if isinstance(timedelta_object, timedelta):
            timedelta_float = timedelta_object.total_seconds()
            timedelta_json = dumps(timedelta_float)
            return timedelta_json
        else:
            error = TypeError(f"Input {timedelta_object} is not a valid datetime.timedelta object.")
            logger.error(error)
            raise error



"""Methods module for qciconnect package. Contains the Method(s) class and related utilities."""
from typeguard import typechecked

from .client import CompilerJob, QciConnectClient
from .exceptions import InvalidIdentifierError, KeycloakTokenManagerError, QciConnectClientError
from .result_handling import CompilerResult, FutureCompilerResult


class Method:
    """Callable object representing a compiler pass."""

    def __init__(
        self, pass_name: str, compiler_id: int, compiler_name: str, client: QciConnectClient
    ):
        """Constructs a callable Method object, which submits a job to respective compile pass.

        Args:
            pass_name: Name of the compiler pass.
            compiler_id: ID of the compiler the compiler pass belongs to.
            compiler_name: Name of the compiler.
            client: Instance of QciConnectClient - QCI Connect RestAPI client.
        """
        self._pass_name = pass_name
        self._compiler_id = compiler_id
        self._compiler_name = compiler_name
        self._compiler_alias = compiler_name.lower().replace(" ", "_").replace("-", "_")
        if not self._compiler_alias.isidentifier():
            raise InvalidIdentifierError(
                f"Compiler alias {self._compiler_alias} is not a valid python identifier."
            )
        self._client = client

    @typechecked
    def __call__(
        self,
        input_circuits: dict[str, str],
        options: dict | None = None,
        wait_for_results: bool = True
    ) -> CompilerResult | FutureCompilerResult | None:
        """Submits input circuits to respective compile pass.

        Args:
            input_circuits: dict of input circuits.
            options: Options determining how to compile.
            wait_for_results: Wait for the computation and return result.
                              Otherwise, return only job ID.

        Returns: If wait_for_results == True, it returns the compilation Result,
                 which is the compiled circuits together with a bunch of meta data.
                 Otherwise, it returns FutureCompilerResult, which promises to return
                  the compilation result when it's available.
        """
        if options is None:
            options = {}

        compiler_job = CompilerJob(self._compiler_id, self._pass_name, options, input_circuits)
        try:
            if wait_for_results:
                result = self._client.submit_compiler_job_and_wait(compiler_job)
                return CompilerResult.from_compiler_task_result(result.last_compiler_result)
            else:
                job_id = self._client.submit_compiler_job(compiler_job)
                return FutureCompilerResult(job_id, self._client)
        except (QciConnectClientError, KeycloakTokenManagerError) as e:
            print(e)
            return None
        except Exception:
            raise

    def __str__(self) -> str:
        """Returns the compiler pass name together with the compiler alias."""
        return f"{self.pass_alias}_{self._compiler_alias}"

    @property
    def pass_alias(self):
        """Returns method/compiler pass alias."""
        return self._pass_name.lower().replace(" ", "_").replace("-", "_")


class Methods:
    """Provides a method/callable object for each compiler pass available on the platform."""

    def __init__(self, compiler_list: list, client: QciConnectClient):
        """Constructs an object containing all Method objects that are supported by the platform.

        This object is constructed by passing a list of compilers and a QciConnectClient.

        Args:
            compiler_list: list of CompilerTable objects.
            client: Instance of QciConnectClient - QCI Connect RestAPI client.
        """
        self._passes = {}
        for compiler in compiler_list:
            for pass_name in compiler["passes_available"]:
                try:
                    method = Method(pass_name, compiler["compiler_id"], compiler["name"], client)
                except InvalidIdentifierError as e:
                    print(e)
                full_pass_name = f"{method.pass_alias}_{method._compiler_alias}"

                if full_pass_name in self._passes:
                    raise InvalidIdentifierError(
                        f"Compiler pass {full_pass_name} already exists."
                    )

                self._passes[full_pass_name] = method

    def __getattr__(self, name) -> Method:
        """Retrieves a compiler pass by name. Raises a KeyError if the pass does not exist."""
        return self._passes.__getitem__(name)

    def __dir__(self):
        """Lists available compiler passes and other attributes/methods."""
        extended_key_list = list(self._passes.keys()) + super().__dir__()
        return extended_key_list

    def show(self):
        """Prints available compiler passes."""
        for method in self._passes.values():
            print(method)



"""Provides a class for interacting with the "QCI Connect" quantum computing platform."""
from typeguard import typechecked

from .client import QciConnectClient
from .exceptions import KeycloakTokenManagerError, QciConnectClientError, ResultError
from .methods import Methods
from .qpus import QpuByAlias
from .result_handling import Results


class Mageia:
    """Represents a frontend to a quantum computing platform.

    Attributes:
        methods: Collects and provides all the methods/compiler passes available on the platform.
        qpus: Collects and provides all the quantum processing units (including simulators)
              available on the platform.
    """

    _DEFAULT_SERVER = "https://qc-plat-pro.dlr.de/"
    _LOCALHOST = "http://127.0.0.1:8000/"

    @typechecked
    def __init__(
        self,
        /,
        token: str | None = None,
        *,
        server: str | None = None,
        username: str | None = None,
        password: str | None = None,
    ):
        """Constructs a Mageia/Client object."""
        if ((username is None or password is None) and token is None) and server is None:
            print(f"Warning! No server and no credentials given.\n"
                  f"In order to use the QCI Connect server {self._DEFAULT_SERVER} "
                  f"you need to provide username and password.\n"
                  f"Falling back to {self._LOCALHOST} as server.")
            server = self._LOCALHOST
        if ((username is not None and password is not None) or token is not None) and \
            server is None:
            server = self._DEFAULT_SERVER

        if username is not None and password is not None:
            print("Deprecation warning: using username and password for authentication will "
                  "be disabled in future. Please create a personal access token on the "
                  f"website {server}")

        try:
            self._client = QciConnectClient(
                server_address=server, 
                username=username, 
                password=password, 
                token=token)
            if server == self._DEFAULT_SERVER:
                self._update_tac_and_privacy_status()
                if not self.tacs_accepted:
                    print(
                        f"Terms and conditions not accepted. Please accept them on "
                        f"{self._DEFAULT_SERVER}."
                    )
                if not self.privacy_notice_accepted:
                    print(
                        f"Privacy notice not accepted. Please accept the privacy notice on "
                        f"{self._DEFAULT_SERVER} using a web browser."
                    )

            self._update_methods()
            self._update_qpus()
            self._update_results()
        except (KeycloakTokenManagerError, QciConnectClientError, ResultError) as e:
            print(e)
        except Exception as e:  # noqa: BLE001
            if (
                server.__contains__("localhost")
                or server.__contains__("0.0.0.0")
                or server.__contains__("127.0.0.1")
            ):
                print(
                    "Please run `docker compose up` before connecting to your backend. \n"
                    "For further information see also https://gitlab.dlr.de/qci-connect/partners/all/qciconnect-sdk"
                )
            else:
                print(f"Connection to {server} could not be established.")
                print(e, type(e))

    @property
    def server_address(self) -> str:
        """Returns URL of the platform."""
        return self._client.server_address

    @server_address.setter
    def server_address(self, value: str):
        """Sets the URL of the platform and retrieves available methods and QPUs.

        Args:
            value: URL of the platform that shall be used.
        """
        self._client.server_address = value
        self._update_methods()
        self._update_qpus()
        self._update_results()

    def _update_methods(self):
        """Updates the methods/compiler passes available on the platform."""
        compiler_list = self._client.get_available_compilers()
        self.methods = Methods(compiler_list, self._client)

    def _update_qpus(self):
        """Updates the quantum processing units available on the platform."""
        self.qpus = QpuByAlias(self._client)

    def _update_results(self):
        """Updates the results of the jobs available on the platform."""
        self.results = Results(self._client)
        self.results.refresh()

    def _update_tac_and_privacy_status(self):
        """Updates the terms and conditions and privacy notice status of the platform."""
        self.tacs_accepted = self._client.get_tac_status()
        self.privacy_notice_accepted = self._client.get_privacy_notice_status()



"""Provides Result object containing result related data for hanling in Celery."""

from datetime import datetime, timedelta
from typing import Any

from qciconnect_conversion.json_conversion import JSONDateTimeConverter, JSONNumpyConverter
from sqlmodel import SQLModel


class ResultForCelery(SQLModel):
    """Contains all result data, meta data that should be fetched by orchestrator.

    Args:
        SQLModel: Class inherits from SQLModel
    """

    data: Any | None
    result_metadata: dict[str, Any] | None
    compiled_circuit: str | None
    start_date: datetime
    end_date: datetime
    status: str | None
    status_message: str | None
    total_run_time: timedelta

    def __init__(self, **data) -> None:
        """Initializes Result with all relevant data."""
        super().__init__(**data)

    def _convert_data(self, data_nparray: Any) -> Any:
        data = JSONNumpyConverter.nparray_to_json_list(data_nparray)
        return data

    def convert_to_json_for_celery(self) -> Any:
        """Converts Result object to JSON.

        Returns:
            JSON: JSON equivalent of result.
        """
        result_dict = {
            "meta_data": self.result_metadata,
            "compiled_circuit": self.compiled_circuit,
            "start_date_time": JSONDateTimeConverter.datetime_to_json_datetime(self.start_date),
            "end_date_time": JSONDateTimeConverter.datetime_to_json_datetime(self.end_date),
            "total_run_time": JSONDateTimeConverter.timedelta_to_json(self.total_run_time),
            "status": self.status,
            "status_message": self.status_message,
            "data": self._convert_data(self.data),
        }
        result_json = JSONNumpyConverter.dict_to_json(result_dict)
        return result_json



"""This module uses https requests to communicate with QCIConnect platform.

Its main class QciConnectClient provides methods to query several REST endpoints.
"""

import json
import time

import httpx
from hequate_common.models import (
    CompilerTaskBase,
    JobResult,
    JobSubmission,
    Primitive,
    QPUOptions,
    QPUTaskBase,
)

from .exceptions import QciConnectClientError
from .token_managers import DummyTokenManager, KeycloakTokenManager, QciConnectTokenManager


class CompilerJob:
    """Class used to hand over the compile jobs to QciConnectClient."""

    def __init__(self, compiler_id: int, compilation_pass: str, options: dict, circuits: dict):
        """Constructor of CompilerJob class."""
        self.task = CompilerTaskBase(
            compiler_id=compiler_id, compilation_pass=compilation_pass, options=options
        )
        self.circuits = circuits


class BackendJob:
    """Class used to hand over the qpu jobs to QciConnectClient."""

    def __init__(
        self,
        backend_id: int,
        circuit: str,
        primitive: str,
        name: str = "Hequate Client Job",
        comment: str = "Issued via API",
        shots: int = 10000,
        qpu_options: dict | None=None,
    ):
        """Initializer of BackendJob class.

        Arguments:
            backend_id: ID that uniquely identifies the backend to be used
            circuit: a string that comprises an OpenQASM 3 circuit.
            primitive: "sampling" or "quantum_state"
            name: Optional name of the job
            comment: Optional string further describing the job
            shots: number of shots
            qpu_options: Additional options for the QPU
        """
        if qpu_options is None:
            qpu_options = {"coupling_map": None, "basis_gates": None}
        self.name = name
        self.comment = comment
        self.circuit = circuit
        self.task = QPUTaskBase(
            qpu_id=backend_id,
            total_shots=shots,
            primitive=Primitive(primitive),
            qpu_options=QPUOptions(**qpu_options),
        )

API_VERSION = "/v1"
class QciConnectClient:
    """Client for QCI Connect frontend RestAPI.

    Handles http get/put/post requests to retrieve available compilers/QPUs
    and to submit jobs to these resources.

    Attributes:
        _server_address: Address of the platform.
        _keycloak_token_manager: Authorization token manager.
    """

    def __init__(self, /, 
        server_address: str, *, 
        username: str | None = None, password: str | None = None, token: str | None = None):
        """Constructs QciConnectClient object.

        Args:
            server_address: Address of the platform.
            username: username for authentication.
            password: password for authentication.
            token: personal access token created on QCI Connect website
        """
        self._username = username
        self._password = password
        self._token = token
        self.server_address = (
            server_address  # setting this also creates a KeycloakTokenManager object
        )

        if isinstance(self._token_manager, DummyTokenManager):
            self._init_endpoints("")
        else:
            self._init_endpoints(API_VERSION)

    @property
    def server_address(self) -> str:
        """Returns the current server address."""
        return self._server_address

    @server_address.setter
    def server_address(self, value: str):
        """Sets the server address.

        Args:
            value: Server address.
        """
        value = value.rstrip("/")
        self._server_address = value
        if self._username is None and self._password is None and self._token is None:
            self._token_manager = DummyTokenManager()
        elif self._username is not None and self._password is not None:
            self._token_manager = KeycloakTokenManager(
                self.server_address, self._username, self._password
            )
        elif self._username is None and self._password is None and self._token is not None:
            self._token_manager = QciConnectTokenManager(self._token)
        elif self._username is not None and \
             self._password is not None and \
             self._token is not None:
             raise QciConnectClientError("Using username/password together with personal "
                "access token is not supprted. If possible, use token only.")
        else:
            raise QciConnectClientError("Error determining the token manager to be used.")

    def _init_endpoints(self, api_version):
        """Initialize API endpoints."""
        self._endpoints = {
            "JOB_LIST_ENDPOINT": "/api" + api_version + "/job/",
            "JOB_SUBMIT_ENDPOINT": "/api" + api_version + "/job/submit",
            "JOB_RESULT_ENDPOINT": "/api" + api_version + "/result/",
            "COMPILER_LIST_ENDPOINT": "/api" + api_version + "/compiler/all",
            "QPU_LIST_ENDPOINT": "/api" + api_version + "/qpu/all",
            "TAC_STATUS_ENDPOINT": "/api" + api_version + "/terms_and_conditions/status",
            "PRIVACY_NOTICE_ENDPOINT": "/api" + api_version + "/privacy_notice/status",
        }

    def _handle_http_status_error(self, errh):
        """Handles HTTP errors.

        Args:
            errh: HTTP error to be handled.
        """
        error_dict = json.loads(errh.response.text)
        if errh.response.status_code == 301:
            # Handle the 301 status code
            raise QciConnectClientError(
                 'Received a 301 status code. '
                f'The requested resource has been permanently moved: "{error_dict["detail"]}"'
            )
        elif errh.response.status_code == 401:
            # Handle the 401 status code
            raise QciConnectClientError(
                 'Received a 401 status code. '
                f'The request was unauthorized: "{error_dict["detail"]}"'
            )
        elif errh.response.status_code == 403:
            # Handle the 403 status code
            raise QciConnectClientError(
                 'Received a 403 status code. '
                f'The request was forbidden: "{error_dict["detail"]}"'
            )
        else:
            # Handle other HTTP errors - print information about the error in a nicely formatted way
            error_detail = error_dict["detail"]
            raise QciConnectClientError(
                f"HTTP Error ({errh.response.status_code}) occurred: {error_detail}"
            )

    def _post(self, request: dict) -> None | str:
        """Sends a post request to the platform and returns the respective job ID.

        Args:
            request: JSON request as a dict.

        Returns: job ID.
        """
        headers = {"Content-Type": "application/json", "Connection": "keep-alive"}
        headers = self._token_manager.add_auth_header(headers)

        with httpx.Client(follow_redirects=False, verify=True, timeout=None) as client:
            try:
                post_response = client.post(
                    f"{self._server_address}{self._endpoints['JOB_SUBMIT_ENDPOINT']}", 
                    json=request,
                    headers=headers
                )
                post_response.raise_for_status()
                job_id = post_response.json()
                print(f"Job submitted with ID: {job_id}")
                return str(job_id)
            except httpx.RequestError as e:
                raise QciConnectClientError(
                    f"A RequestError occurred while sending a post-request to URL: "
                    f"{e.request.url} - {e}"
                ) from e
            except httpx.HTTPStatusError as e:
                self._handle_http_status_error(e)

    def _job_id_to_result_endpoint(self, job_id: str) -> str:
        """Converts a job ID to the respective result endpoint.

        Args:
            job_id: Job ID to be converted.

        Returns: URL to the respective result endpoint.
        """
        return f"{self._server_address}{self._endpoints['JOB_RESULT_ENDPOINT']}{job_id}"

    def _job_id_to_job_endpoint(self, job_id: str) -> str:
        """Converts a job ID to the respective result endpoint.

        Args:
            job_id: Job ID to be converted.

        Returns: URL to the respective result endpoint.
        """
        return f"{self._server_address}{self._endpoints['JOB_LIST_ENDPOINT']}{job_id}"

    def _get(self, get_url: str) -> dict | list[dict]:
        """Sends a get request for a given URL.

        Args:
            get_url: URL to which a get request is sent.

        Returns: JSON response in form of a dict.
        """
        headers = {"Content-Type": "application/json", "Connection": "keep-alive"}
        headers = self._token_manager.add_auth_header(headers)

        with httpx.Client(follow_redirects=False, verify=True, timeout=None) as client:
            try:
                get_response = client.get(get_url, headers=headers)
                get_response.raise_for_status()
                payload = get_response.json()
                return payload # noqa: TRY300
            except httpx.HTTPStatusError as errh:
                self._handle_http_status_error(errh)
            except httpx.RequestError as e:
                raise QciConnectClientError(
                    f"A RequestError occurred while sending a get-request to URL: "
                    f"{e.request.url} - {e}."
                ) from e
            except Exception as e:
                # Handle other exceptions
                raise QciConnectClientError(f"An error occurred: {e}") from e

    def _sanitize_circuit(self, circuit: str) -> str:
        """Sanitizes the circuit.

        Args:
            circuit: Circuit to be sanitized.

        Returns: Sanitized circuit.
        """
        if type(circuit) is not str:
            raise QciConnectClientError("Circuit must be a string.")

        circuit = circuit.replace("OPENQASM 2.0;", "")
        circuit = circuit.replace('include "qelib1.inc";', 'include "stdgates.inc";')

        return circuit

    def submit_compiler_job(self, compiler_job: CompilerJob):
        """Submits a compiler job.

        Args:
            compiler_job: Quantum circuit(s) and information about the compiler pass to be used.

        Returns: id identifying the job just submitted.
        """
        circuit = next(iter(compiler_job.circuits.values()))
        circuit = self._sanitize_circuit(circuit)

        request = JobSubmission(
            circuit=circuit,
            name="Hequate Client Compile Job",
            comment="Issued via API",
            tasks=[compiler_job.task],
        )

        job_id = self._post(request.model_dump())
        return job_id

    def _has_job_finished(self, response: dict) -> bool:
        """Interprets the status message.

        Args:
            response: dictionary that contains status message to be interpreted.
        """
        if response["status"] == "SUCCESS":
            print("Job was successful.")
            return True
        elif response["status"] == "FAILURE":
            try:
                print(f"{response['status_message']}")
            except KeyError:
                print("Job failed.")
            return True
        else:
            print("Job is not finished yet.")
            return False

    def get_job_status(self, job_id: str) -> str:
        """Gets the status of a job.

        Args:
            job_id: id of job which status shall be returned.

        Returns: string telling the status of the job.
        """
        endpoint_url = self._job_id_to_result_endpoint(job_id)
        response = self._get(endpoint_url)
        return response["status"]

    def get_job(self, job_id: str) -> dict:
        """Queries the job endpoint to return the info on a specific job.

        Args:
            job_id: UID that defines the job to be queried.

        Returns:
            dict with information on the given job
        """
        endpoint_url = self._job_id_to_job_endpoint(job_id)
        return self._get(endpoint_url)

    def retrieve_result(self, job_id: str) -> JobResult | None:
        """Retrieves the result of a job by its job ID.

        Args:
            job_id: string with the job's UID

        Returns:
            JobResult object
        """
        endpoint_url = self._job_id_to_result_endpoint(job_id)
        response = self._get(endpoint_url)
        if self._has_job_finished(response):
            result = JobResult.model_validate(response)
            return result
        else:
            return None

    def wait_and_call_method(self, timeout, method, *params):
        """Calls `method` every 5 seconds until it returns not None or timeout is exceeded.

        Args:
            timeout: timeout in seconds. An exception is raised when the `method` does not
                     return a value in that time.
            method: the function or method to be called
            params: parameters for the function or method to be called.

        """
        start_time = time.time()
        while time.time() - start_time < timeout:
            result = method(*params)
            if result is not None:
                return result
            time.sleep(5)
        raise QciConnectClientError("Timeout occurred while waiting for job result.")

    def submit_compiler_job_and_wait(self, compiler_job: CompilerJob, timeout=3600) -> JobResult:
        """Submits a compiler job and waits for the result.

        Args:
            compiler_job: Quantum circuit(s) and information about the compiler pass to be used.
            timeout: Timeout in seconds.

        Returns: JobResult which is compiled quantum circuit(s) and a bunch of meta information.
        """
        job_id = self.submit_compiler_job(compiler_job)
        result = self.wait_and_call_method(timeout, self.retrieve_result, job_id)
        return result

    def submit_backend_job(self, backend_job: BackendJob) -> str:
        """Submits a quantum circuit processing job.

        Args:
            backend_job: Quantum circuit and information about the QPU it should run on (and how).

        Returns: id identifying the job just submitted.
        """
        circuit = self._sanitize_circuit(backend_job.circuit)

        request = JobSubmission(
            circuit=circuit,
            primitive=backend_job.task.primitive,
            name=backend_job.name,
            comment=backend_job.comment,
            tasks=[backend_job.task],
        )

        job_id = self._post(request.model_dump())
        return job_id

    def submit_backend_job_and_wait(
        self, backend_job: BackendJob, timeout: int = 3600
    ) -> JobResult:
        """Submits a quantum circuit processing job and waits for the result.

        Args:
            backend_job: Quantum circuit and information about the QPU it should run on (and how).
            timeout: Timeout in seconds.

        Returns: JobResult which is measurement data and a bunch of meta information.
        """
        job_id = self.submit_backend_job(backend_job)
        result = self.wait_and_call_method(timeout, self.retrieve_result, job_id)
        return result

    def get_available_compilers(self) -> list:
        """Retrieves list of compilers available on the platform."""
        endpoint_url = f"{self._server_address}{self._endpoints['COMPILER_LIST_ENDPOINT']}"
        return self._get(endpoint_url)

    def get_available_qpus(self) -> list[dict]:
        """Retrieves list of QPUs available on the platform."""
        endpoint_url = f"{self._server_address}{self._endpoints['QPU_LIST_ENDPOINT']}"
        return self._get(endpoint_url)

    def get_jobs(self) -> list[dict]:
        """Retrieves list of jobs available in the platform database."""
        endpoint_url = f"{self._server_address}{self._endpoints['JOB_LIST_ENDPOINT']}"
        return self._get(endpoint_url)

    def get_tac_status(self) -> bool:
        """Retrieves status of terms and conditions acceptance."""
        endpoint_url = f"{self._server_address}{self._endpoints['TAC_STATUS_ENDPOINT']}"
        return self._get(endpoint_url)

    def get_privacy_notice_status(self) -> bool:
        """Retrieves status of privacy notice acceptance."""
        endpoint_url = f"{self._server_address}{self._endpoints['PRIVACY_NOTICE_ENDPOINT']}"
        return self._get(endpoint_url)



"""QPU and QpusByAlias classes for qciconnect package."""
from pydantic import ValidationError as PydanticValidationError
from tabulate import tabulate
from typeguard import typechecked

from qciconnect.exceptions import QciConnectClientError

from .client import BackendJob, QciConnectClient
from .result_handling import BackendResult, FutureBackendResult


class Qpu:
    """Class representing a quantum circuit processing unit (real device or simulator)."""

    def __init__(
        self,
        identifier: int,
        name: str,
        manufacturer: str,
        qubit_count: int,
        status: str,
        client: QciConnectClient,
    ):
        """Constructs a Qpu object.

        Args:
            identifier: Unique identifier of the quantum circuit processing unit.
            name: Name of the quantum circuit processing unit.
            manufacturer: Manufacturer of the quantum circuit processing unit.
            qubit_count: Number of qubits available on the QPU.
            status: Current operational status.
            client: Instance of QciConnectClient - QCI Connect RestAPI client.
        """
        self._id = identifier
        self._name = name
        self._alias = name.lower().replace(" ", "_").replace("-", "_")
        self._manufacturer = manufacturer
        self._qubit_count = qubit_count
        self._status = status
        self._client = client

    @typechecked
    def submit(
        self,
        circuit: str,
        primitive: str,
        shots: int = 10000,
        wait_for_results: bool = True,
        name: str = "Hequate QPU Job",
        comment: str = "Issued via API",
        qpu_options: dict | None = None,
    ) -> FutureBackendResult | BackendResult | None:
        """Submits the circuit to the QPU for execution according to primitive and shots.

        Args:
            name: Name of the job.
            circuit: Quantum circuit to be executed.
            primitive: Way of execution.
            shots: Number of runs/executions.
            wait_for_results: wait for the backend to finished the submitted job.
            comment: Optional string to further describe the job
            qpu_options: Additional options for the QPU.

        Returns: BackendResult, which is measurement data with a bunch of meta data or
                 FutureBackendResult which is a promise to later return a BackendResult
        """
        if qpu_options is None:
            qpu_options = {}
        try:
            job = BackendJob(
                self._id,
                circuit,
                primitive,
                name=name,
                shots=shots,
                comment=comment,
                qpu_options=qpu_options
            )
        except (PydanticValidationError, ValueError) as e:
            print(e)
            return None

        if wait_for_results:
            try:
                result = self._client.submit_backend_job_and_wait(job)
                if result.last_qpu_result is None:
                    print("Job failed(?) - no result data available.")
                    return None
            except (QciConnectClientError, PydanticValidationError) as e:
                print(e)
                return None
            return BackendResult.from_qpu_task_result(result.last_qpu_result)
        else:
            try:
                job_id = self._client.submit_backend_job(job)
            except (QciConnectClientError, PydanticValidationError) as e:
                print(e)
                return None
            return FutureBackendResult(job_id, self._client)

    def __str__(self) -> str:
        """Returns a string representation of the QPU with its alias, qubit count, and status."""
        return f"Name: {self._alias}, #Qubits: {self._qubit_count}, Status: {self._status}"

    def get_qpu_info(self) -> list[str]:
        """Returns list containing QPU alias, #qubits, and status."""
        return [self._alias, str(self._qubit_count), self._status]

    def __dir__(self):
        """Returns a list of attributes and methods of the specific Qpu  (w/o dunders)."""
        method_list = []
        for attr in dir(self):
            if not attr.startswith("__"):
                method_list.append(attr)
        return method_list


class QpuByAlias:
    """Dictionary of available QPUs on the platform indexed by their aliases."""

    def __init__(self, client: QciConnectClient):
        """Constructs a dict of QPUs available on the platform (indexed by their aliases).

        Args:
            qpu_list: list of QPUTable objects.
            client: Instance of QciConnectClient - QCI Connect RestAPI client.
        """
        self._client = client
        self._update_qpu_dict()

    def _update_qpu_dict(self):
        qpu_list = self._client.get_available_qpus()
        self._qpus = {}
        for qpu_entry in qpu_list:
            qpu = Qpu(
                qpu_entry["qpu_id"],
                qpu_entry["name"],
                qpu_entry["manufacturer"],
                qpu_entry["number_of_qubits_available"],
                qpu_entry["status"],
                self._client,
            )
            self._qpus[qpu._alias] = qpu

    def __getattr__(self, name) -> Qpu | None:
        """Returns a Qpu object if it exists, otherwise None."""
        try:
            return self._qpus.__getitem__(name)
        except KeyError as e:
            print(e)
            return None

    def __dir__(self):
        """Returns a list of QPUs and other attributes."""
        extended_key_list = list(self._qpus.keys()) + list(super().__dir__())
        return extended_key_list

    def show(self):
        """Prints table of available QPUs including their qubit counts and status."""
        self._update_qpu_dict()
        all_qpu_info = []
        for qpu in self._qpus.values():
            all_qpu_info.append(qpu.get_qpu_info())

        print(tabulate(all_qpu_info, headers=["Alias", "Qubits", "Status"]))
